#!/usr/bin/env python3

"""
This script modifies the ARGannot_clustered80_r3.csv file (comes with SRST2) for use with
Kleborate. Specifically, it does two things:
  1) Add beta-lactamase description and class columns
  2) Removes some genes not suitable for inclusion in Kleborate

It takes two arguments:
  1) SRST2's copy of the ARGannot_clustered80_r3.csv file
  2) the table made by bla_info.py

Usage:
  ./make_argannot_csv.py path/to/srst2/data/ARGannot_clustered80_r3.csv bla_info_table > ARGannot_clustered80_r3.csv
"""


import sys

argannot_csv_filename = sys.argv[1]
bla_parse_output_filename = sys.argv[2]

descriptions = {}
bla_classes = {}

with open(bla_parse_output_filename, 'rt') as bla_parse_output_file:
    for line in bla_parse_output_file:
        allele, description, bla_class = line.strip().split('\t')
        descriptions[allele] = description
        bla_classes[allele] = bla_class

new_csv_lines = []
with open(argannot_csv_filename, 'rt') as argannot_csv_file:
    for line in argannot_csv_file:
        stripped_line = line.rstrip()

        if stripped_line.startswith('clusterid,queryID'):
            print(stripped_line + ',bla_description,bla_class')
            continue

        # Fix verbose gene name
        stripped_line = stripped_line.replace('NimB_Nitroimidazole_Gene', 'NimB')

        # Fix resistance class capitalisation
        stripped_line = stripped_line.replace(',Agly,', ',AGly,')

        line_parts = stripped_line.split(',')
        res_class = line_parts[2]
        gene = line_parts[3]
        allele = line_parts[4]

        # Skip some genes not relevant for Kleborate.
        if gene == 'OqxB' or gene == 'OqxA' or gene == 'Far1' or gene == 'OptrA':
            continue

        # Fix Colistin class name to just three letters.
        if res_class == 'Colistin':
            stripped_line = stripped_line.replace(',Colistin,', ',Col,')

        if line_parts[2] == 'Bla' and allele in bla_classes:
            new_csv_lines.append(stripped_line + ',' + descriptions[allele] + ',' + bla_classes[allele])
        else:
            new_csv_lines.append(stripped_line + ',NA,NA')

# Sort so 'TEM-2' will come before 'TEM-10', etc.
new_csv_lines = sorted(new_csv_lines, key=lambda x: (int(x.split(',')[0]), int(x.split(',')[5])))

for line in new_csv_lines:
    print(line)




"""
This module contains some subclasses for customising Trycycler's help text formatting.

Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/rrwick/Trycycler

This file is part of Trycycler. Trycycler is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Trycycler is distributed
in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Trycycler.
If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import os
import shutil
import subprocess
import sys


END_FORMATTING = '\033[0m'
BOLD = '\033[1m'
DIM = '\033[2m'


class MyParser(argparse.ArgumentParser):
    """
    This subclass of ArgumentParser changes the error messages, such that if a command is run with
    no other arguments, it will display the help text. If there is a different error, it will give
    the normal response (usage and error).
    """
    def error(self, message):
        if len(sys.argv) == 2:  # if a command was given but nothing else
            self.print_help(file=sys.stderr)
            sys.exit(2)
        else:
            super().error(message)


class MyHelpFormatter(argparse.HelpFormatter):
    """
    This is a custom formatter class for argparse. It allows for some custom formatting,
    in particular for the help texts with multiple options (like bridging mode and verbosity level).
    http://stackoverflow.com/questions/3853722
    """
    def __init__(self, prog):
        terminal_width = shutil.get_terminal_size().columns
        os.environ['COLUMNS'] = str(terminal_width)
        max_help_position = min(max(24, terminal_width // 3), 40)
        self.colours = get_colours_from_tput()
        super().__init__(prog, max_help_position=max_help_position)

    def _get_help_string(self, action):
        """
        Override this function to add default values, but only when 'default' is not already in the
        help text.
        """
        help_text = action.help
        if action.default != argparse.SUPPRESS and action.default is not None:
            if 'NO_DEFAULT' in help_text:
                help_text = help_text.replace('NO_DEFAULT', '').strip()
            elif 'default' not in help_text.lower():
                help_text += ' (default: {})'.format(action.default)
            elif 'default: DEFAULT' in help_text:
                help_text = help_text.replace('default: DEFAULT',
                                              'default: {}'.format(action.default))
        return help_text

    def start_section(self, heading):
        """
        Override this method to add bold underlining to section headers.
        """
        if self.colours > 1:
            heading = BOLD + heading + END_FORMATTING
        super().start_section(heading)

    def _split_lines(self, text, width):
        """
        Override this method to add special behaviour for help texts that start with:
          'R|' - loop text one option per line
        """
        if text.startswith('R|'):
            text_lines = text[2:].splitlines()
            wrapped_text_lines = []
            for line in text_lines:
                if len(line) <= width:
                    wrapped_text_lines.append(line)
                else:
                    wrap_column = 2
                    line_parts = line.split(', ')
                    join = ','
                    current_line = line_parts[0]
                    for part in line_parts[1:]:
                        if len(current_line) + len(join) + 1 + len(part) <= width:
                            current_line += join + ' ' + part
                        else:
                            wrapped_text_lines.append(current_line + join)
                            current_line = ' ' * wrap_column + part
                    wrapped_text_lines.append(current_line)
            return wrapped_text_lines
        else:
            return argparse.HelpFormatter._split_lines(self, text, width)

    def _fill_text(self, text, width, indent):
        if text.startswith('R|'):
            return ''.join(indent + line for line in text[2:].splitlines(keepends=True))
        else:
            return argparse.HelpFormatter._fill_text(self, text, width, indent)

    def _format_action(self, action):
        """
        Override this method to make help descriptions dim.
        """
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2,
                            self._max_help_position)
        help_width = self._width - help_position
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)

        # ho nelp; start on same line and add a final newline
        if not action.help:
            tup = self._current_indent, '', action_header
            action_header = '%*s%s\n' % tup
            indent_first = 0

        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            tup = self._current_indent, '', action_width, action_header
            action_header = '%*s%-*s  ' % tup
            indent_first = 0

        # long action name; start on the next line
        else:
            tup = self._current_indent, '', action_header
            action_header = '%*s%s\n' % tup
            indent_first = help_position

        # collect the pieces of the action help
        parts = [action_header]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            first_line = help_lines[0]
            if self.colours > 8:
                first_line = DIM + first_line + END_FORMATTING
            parts.append('%*s%s\n' % (indent_first, '', first_line))
            for line in help_lines[1:]:
                if self.colours > 8:
                    line = DIM + line + END_FORMATTING
                parts.append('%*s%s\n' % (help_position, '', line))

        # or add a newline if the description doesn't end with one
        elif not action_header.endswith('\n'):
            parts.append('\n')

        # if there are any sub-actions, add their help as well
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))

        # return a single string
        return self._join_parts(parts)


def get_colours_from_tput():
    try:
        return int(subprocess.check_output(['tput', 'colors']).decode().strip())
    except (ValueError, subprocess.CalledProcessError, FileNotFoundError, AttributeError):
        return 1



"""
Blast for resistance genes, summarise by class (one class per column)

Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections

from Bio import pairwise2
from Bio.Align import substitution_matrices
from Bio.Seq import Seq
from Bio.Data.CodonTable import TranslationError

from .blastn import run_blastn
from .misc import load_fasta, reverse_complement
from .shv_mutations import check_for_shv_mutations
from .truncation import truncation_check


def resblast_one_assembly(contigs, gene_info, qrdr, trunc, omp, seqs, min_cov, min_ident,
                          min_spurious_cov, min_spurious_ident):
    hits_dict = blast_against_all(seqs, min_cov, min_ident, contigs, gene_info,
                                  min_spurious_cov, min_spurious_ident)
    if qrdr:
        check_for_qrdr_mutations(hits_dict, contigs, qrdr, min_ident, 90.0)
    if trunc:
        check_for_mgrb_pmrb_gene_truncations(hits_dict, contigs, trunc, min_ident)
    if omp:
        check_omp_genes(hits_dict, contigs, omp, min_ident, 90.0)
    return hits_dict


def read_class_file(res_class_file):
    gene_info = {}  # key = sequence id (fasta header in seq file), value = (allele,class,Bla_Class)
    res_classes = []
    bla_classes = ['Bla', 'Bla_inhR', 'Bla_ESBL', 'Bla_ESBL_inhR', 'Bla_Carb', 'Bla_chr']

    with open(res_class_file, 'r') as f:
        header = 0
        for line in f:
            if header == 0:
                header = 1
                # clusterid,queryID,class,gene,allele,seqID,accession,positions,size,
                # cluster_contains_multiple_genes,gene_found_in_multiple_clusters,bla_description,
                # bla_class
            else:
                fields = line.rstrip().split(',')

                cluster_id, res_class, gene, allele_symbol, seq_id, bla_class = \
                    fields[0], fields[2], fields[3], fields[4], fields[5], fields[12]
                seq_header = '__'.join([cluster_id, gene + '_' + res_class, allele_symbol, seq_id])

                if res_class == 'Bla' and bla_class == 'NA':
                    bla_class = 'Bla'
                gene_info[seq_header] = (allele_symbol, res_class, bla_class)
                if res_class not in res_classes:
                    res_classes.append(res_class)
                if bla_class not in bla_classes:
                    bla_classes.append(bla_class)

    res_classes.sort()
    if 'Bla' in res_classes:
        res_classes.remove('Bla')
    if 'NA' in bla_classes:
        bla_classes.remove('NA')

    if 'SHV_mutations' not in res_classes:
        res_classes.append('SHV_mutations')
    if 'Omp_mutations' not in res_classes:
        res_classes.append('Omp_mutations')
    if 'Col_mutations' not in res_classes:
        res_classes.append('Col_mutations')
    if 'Flq_mutations' not in res_classes:
        res_classes.append('Flq_mutations')

    return gene_info, res_classes, bla_classes


def get_res_headers(res_classes, bla_classes):
    res_headers = res_classes + bla_classes

    # Rearrange the headers a bit. First move Bla_chr to the end:
    res_headers = ([h for h in res_headers if h != 'Bla_chr'] +
                   [h for h in res_headers if h == 'Bla_chr'])

    # Then move mutation columns to the end:
    res_headers = ([h for h in res_headers if '_mutations' not in h] +
                   [h for h in res_headers if '_mutations' in h])

    # Add '_acquired' to the end of the rest of the columns:
    res_headers = [h if h.endswith('_chr') or h.endswith('_mutations') else h + '_acquired'
                   for h in res_headers]

    return res_headers


def blast_against_all(seqs, min_cov, min_ident, contigs, gene_info, min_spurious_cov,
                      min_spurious_ident):
    hits_dict = collections.defaultdict(list)  # key = class, value = list
    hits = run_blastn(seqs, contigs, min_spurious_cov, min_spurious_ident)
    for hit in hits:
        coverage = hit.alignment_length / hit.ref_length * 100.0
        if coverage >= min_spurious_cov:
            if hit.pcid < 100.0:
                aa_result = check_for_exact_aa_match(seqs, hit, contigs)
                if aa_result is not None:
                    hit.gene_id = aa_result
                    exact_match = True
                else:
                    exact_match = False
            else:
                aa_result = None
                exact_match = True

            hit_allele, hit_class, hit_bla_class = gene_info[hit.gene_id]

            hit_bla_class, shv_muts, class_changing_muts, omega_loop_seq = \
                check_for_shv_mutations(hit, hit_allele, hit_bla_class, exact_match)

            if hit_class == 'Bla':
                hit_class = hit_bla_class

            hits_dict['SHV_mutations'] += shv_muts
            if omega_loop_seq is not None:
                hits_dict['SHV_mutations'].append(f'omega-loop={omega_loop_seq}')
            if not hits_dict['SHV_mutations']:
                del hits_dict['SHV_mutations']

            if not (hit_class.endswith('_chr') or hit_class.endswith('_mutations')):
                hit_class += '_acquired'

            trunc_cov = 100.0
            if aa_result is not None:
                hit_allele += '^'
            else:
                if hit.pcid < 100.0:
                    hit_allele += '*'
                if hit.alignment_length < hit.ref_length:
                    hit_allele += '?'
                trunc_suffix, trunc_cov, _ = truncation_check(hit)
                hit_allele += trunc_suffix

            if class_changing_muts:
                hit_allele += ' +' + ' +'.join(class_changing_muts)

            # If the hit is decent (above the min coverage and identity thresholds), it goes in the
            # column for the class.
            if coverage >= min_cov and hit.pcid >= min_ident and trunc_cov >= 90.0:
                hits_dict[hit_class].append(hit_allele)

            # If the hit is decent but the gene is truncated, it goes in the
            # truncated_resistance_hits column.
            elif coverage >= min_cov and hit.pcid >= min_ident and trunc_cov < 90.0:
                hits_dict['truncated_resistance_hits'].append(hit_allele)

            # If the hit is bad (below the min coverage and identity thresholds but above the
            # thresholds for spurious hits) then it goes in the spurious hit column.
            else:
                hits_dict['spurious_resistance_hits'].append(hit_allele)

    return hits_dict


def check_for_exact_aa_match(seqs, hit, contigs):
    """
    This function checks to see if an exact amino acid match can be found for a sequence that had
    an inexact nucleotide match. If so, return the gene_id, otherwise None. If multiple references
    have exact amino acid matches, it returns the longest one. If multiple references have
    equally-long exact amino acid matches, it returns the alphabetically first.
    """
    # First, we extract the nucleotide sequence from the assembly.
    hit_seq, _, _ = hit.get_seq_start_end_pos_strand()
    assembly_seqs = dict(load_fasta(contigs))
    contig_start, contig_end = hit.contig_start-1, hit.contig_end  # 1-based to 0-based indexing
    contig_length = len(assembly_seqs[hit.contig_name])
    gene_nucl_seq = assembly_seqs[hit.contig_name][contig_start:contig_end]
    if hit.strand == 'minus':
        gene_nucl_seq = reverse_complement(gene_nucl_seq)
    assert hit_seq == gene_nucl_seq

    # We also need to check whether the first few or last few bases of the sequence is missing.
    # This is to catch cases where an alternative start/stop codon can lead to an incomplete
    # nucleotide match even when there is an exact amino acid match. If we find that the hit is
    # missing start or end bases (relative to the reference), then we add those back on and will
    # include this augmented sequence in the exact amino acid check.
    ref_seqs = load_fasta(seqs)
    ref_length = len(dict(ref_seqs)[hit.gene_id])
    ref_start, ref_end = sorted([hit.ref_start, hit.ref_end])
    ref_start -= 1  # 1-based to 0-based indexing
    missing_start = ref_start
    missing_end = ref_length - ref_end
    if missing_start == 0 and missing_end == 0:
        augmented_gene_nucl_seq = None
    elif missing_start > 10 and missing_end > 10:  # don't bother with too much missing start/end
        augmented_gene_nucl_seq = None
    else:
        if hit.strand == 'plus':
            contig_start -= missing_start
            contig_end += missing_end
        elif hit.strand == 'minus':
            contig_start -= missing_end
            contig_end += missing_start
        else:
            assert False
        contig_start = max(contig_start, 0)
        contig_end = min(contig_end, contig_length)
        augmented_gene_nucl_seq = assembly_seqs[hit.contig_name][contig_start:contig_end]
        if hit.strand == 'minus':
            augmented_gene_nucl_seq = reverse_complement(augmented_gene_nucl_seq)

    # Look for an amino acid match between the assembly sequence and any reference sequence.
    best_match_length = 0
    best_matches = []
    for name, ref_nucl_seq in ref_seqs:
        match = is_exact_aa_match(gene_nucl_seq, ref_nucl_seq)
        if augmented_gene_nucl_seq is not None and \
                is_exact_aa_match(augmented_gene_nucl_seq, ref_nucl_seq):
            match = True
        if match:
            if len(ref_nucl_seq) > best_match_length:
                best_matches = [name]
                best_match_length = len(ref_nucl_seq)
            elif len(ref_nucl_seq) == best_match_length:
                best_matches.append(name)
    if not best_matches:
        return None
    else:
        return sorted(best_matches)[0]


def is_exact_aa_match(gene_nucl_seq_1, ref_nucl_seq):
    # We look at the gene nucleotide sequence in all three frames of the forward strand.
    gene_nucl_seq_2 = gene_nucl_seq_1[1:]
    gene_nucl_seq_3 = gene_nucl_seq_1[2:]

    gene_prot_1 = translate_nucl_to_prot(gene_nucl_seq_1)
    gene_prot_2 = translate_nucl_to_prot(gene_nucl_seq_2)
    gene_prot_3 = translate_nucl_to_prot(gene_nucl_seq_3)
    ref_prot = translate_nucl_to_prot(ref_nucl_seq)

    # If the reference protein sequence is contained within any frame of the gene protein sequence,
    # that counts as a match.
    return (ref_prot in gene_prot_1) or (ref_prot in gene_prot_2) or (ref_prot in gene_prot_3)


def translate_nucl_to_prot(nucl_seq):
    # First try to translate as a complete coding sequence. This will allow for alternative start
    # codons (e.g. GTG -> M) if it works. We have to manually add the stop codon (*) here because
    # using cds=True turns that off.
    try:
        return str(Seq(nucl_seq).translate(table='Bacterial', to_stop=False, cds=True)) + '*'
    except TranslationError:
        pass

    # If that failed, we will translate in a more relaxed way using a nucleotide sequence truncated
    # to a multiple-of-three length.
    truncated_nucl_seq = nucl_seq[:len(nucl_seq) // 3 * 3]
    return str(Seq(truncated_nucl_seq).translate(table='Bacterial', to_stop=False, cds=False))


def check_for_qrdr_mutations(hits_dict, contigs, qrdr, min_ident, min_cov):
    qrdr_loci = {'GyrA': [(83, 'S'), (87, 'D')],
                 'ParC': [(80, 'S'), (84, 'E')]}

    gyra_ref = 'MSDLAREITPVNIEEELKNSYLDYAMSVIVGRALPDVRDGLKPVHRRVLYAMNVLGNDWN' \
               'KAYKKSARVVGDVIGKYHPHGDSAVYDTIVRMAQPFSLRYMLVDGQGNFGSIDGDSAAAM'
    parc_ref = 'MSDMAERLALHEFTENAYLNYSMYVIMDRALPFIGDGLKPVQRRIVYAMSELGLNASAKF' \
               'KKSARTVGDVLGKYHPHGDSACYEAMVLMAQPFSYRYPLVDGQGNWGAPDDPKSFAAMRY'

    blosum62 = substitution_matrices.load('BLOSUM62')

    snps = []
    hits = run_blastn(qrdr, contigs, None, min_ident)
    for hit in hits:
        _, coverage, translation = truncation_check(hit)
        if coverage > min_cov:
            if hit.gene_id == 'GyrA':
                alignments = pairwise2.align.globalds(gyra_ref, translation, blosum62, -10, -0.5)
            elif hit.gene_id == 'ParC':
                alignments = pairwise2.align.globalds(parc_ref, translation, blosum62, -10, -0.5)
            else:
                assert False
            bases_per_ref_pos = get_bases_per_ref_pos(alignments[0])
            loci = qrdr_loci[hit.gene_id]
            for pos, wt_base in loci:
                assembly_base = bases_per_ref_pos[pos]
                if pos in bases_per_ref_pos and assembly_base != wt_base and \
                        assembly_base != '-' and assembly_base != '.':
                    snps.append(hit.gene_id + '-' + str(pos) + assembly_base)
    if snps:
        hits_dict['Flq_mutations'] += snps


def get_bases_per_ref_pos(alignment):
    aligned_seq1, aligned_seq2 = alignment[0], alignment[1]
    bases_per_ref_pos = {}
    ref_pos = 1
    for i, ref_b in enumerate(aligned_seq1):
        if ref_b == '-' or ref_b == '.':
            continue
        assembly_b = aligned_seq2[i]
        bases_per_ref_pos[ref_pos] = assembly_b
        ref_pos += 1
    return bases_per_ref_pos


def check_for_mgrb_pmrb_gene_truncations(hits_dict, contigs, seqs, min_ident):
    best_mgrb_cov, best_pmrb_cov = 0.0, 0.0

    hits = run_blastn(seqs, contigs, None, min_ident)
    for hit in hits:
        assert hit.gene_id == 'pmrB' or hit.gene_id == 'mgrB'
        _, coverage, _ = truncation_check(hit)
        if hit.gene_id == 'mgrB' and coverage > best_mgrb_cov:
            best_mgrb_cov = coverage
        elif hit.gene_id == 'pmrB' and coverage > best_pmrb_cov:
            best_pmrb_cov = coverage

    truncations = []
    if best_mgrb_cov < 90.0:
        truncations.append('MgrB-' + ('%.0f' % best_mgrb_cov) + '%')
    if best_pmrb_cov < 90.0:
        truncations.append('PmrB-' + ('%.0f' % best_pmrb_cov) + '%')

    if truncations:
        hits_dict['Col_mutations'] += truncations


def check_omp_genes(hits_dict, contigs, omp, min_ident, min_cov):
    best_ompk35_cov, best_ompk36_cov = 0.0, 0.0

    hits = run_blastn(omp, contigs, None, min_ident)
    for hit in hits:
        _, coverage, translation = truncation_check(hit)
        if hit.gene_id == 'OmpK35':
            if coverage > best_ompk35_cov:
                best_ompk35_cov = coverage
        elif hit.gene_id == 'OmpK36':
            if coverage > best_ompk36_cov:
                best_ompk36_cov = coverage
            if coverage >= min_cov:
                if 'GDGDTY' in translation:
                    hits_dict['Omp_mutations'].append('OmpK36GD')
                elif 'GDTDTY' in translation:
                    hits_dict['Omp_mutations'].append('OmpK36TD')
        else:
            assert False

    truncations = []
    if best_ompk35_cov < 90.0:
        truncations.append('OmpK35-' + ('%.0f' % best_ompk35_cov) + '%')
    if best_ompk36_cov < 90.0:
        truncations.append('OmpK36-' + ('%.0f' % best_ompk36_cov) + '%')

    if truncations:
        if 'Omp_mutations' not in hits_dict:
            hits_dict['Omp_mutations'] = []
        hits_dict['Omp_mutations'] += truncations



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

from Bio.Seq import Seq


def truncation_check(hit, cov_threshold=90.0):
    """
    Checks to see if the gene is truncated at the amino acid level.
    """
    nucl_seq, ref_start, _ = hit.get_seq_start_end_pos_strand()

    # The hit must start at the first base of the gene. If not, the gene is considered 0%.
    if ref_start != 1:
        return '-0%', 0.0, ''

    # If there are any ambiguous bases in the sequence, then they will break translation, probably
    # resulting in truncation call.
    ambiguous_bases = set(b for b in nucl_seq) - {'A', 'C', 'G', 'T'}
    for b in ambiguous_bases:
        nucl_seq = nucl_seq.split(b)[0]

    # BioPython doesn't like it if the sequence isn't a multiple of 3.
    nucl_seq = nucl_seq[:len(nucl_seq) // 3 * 3]

    # The assumption is that the reference allele is a full CDS with a stop codon at the end. This
    # isn't always true (the reference sequence is sometimes broken) but will serve to make our
    # denominator for coverage.
    ref_aa_length = (hit.ref_length - 3) // 3

    coding_dna = Seq(nucl_seq)
    translation = str(coding_dna.translate(table='Bacterial', to_stop=True))

    coverage = 100.0 * len(translation) / ref_aa_length
    if coverage >= cov_threshold:
        return '', coverage, translation
    else:
        return '-{:.0f}%'.format(coverage), coverage, translation



#!/usr/bin/env python3
"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import subprocess
import distutils.spawn
from setuptools import setup
from setuptools.command.install import install


def readme():
    with open('README.md') as f:
        return f.read()


# Get the program version from another file.
__version__ = '0.0.0'
exec(open('kleborate/version.py').read())


def check_dir_write_permission(directory):
    if os.path.isdir(directory) and not os.access(directory, os.W_OK):
        sys.exit('Error: no write permission for ' + directory + '  ' +
                 'Perhaps you need to use sudo?')


def build_blast_db(data_dir, fasta_filename, seq_type):
    fasta_path = os.path.join(data_dir, fasta_filename)
    makeblastdb_cmd = ['makeblastdb', '-dbtype', seq_type, '-in', fasta_path]
    print('  ' + ' '.join(makeblastdb_cmd))
    with open(os.devnull, 'w') as devnull:
        subprocess.check_call(makeblastdb_cmd, stdout=devnull)


class KleborateInstall(install):
    """
    This is a subclass of install where I've added a couple things to the installation process:
      * building BLAST databases with makeblastdb
      * printing example commands
    """

    def run(self):
        check_dir_write_permission(self.install_lib)
        check_dir_write_permission(self.install_scripts)

        # Call the base class run to do the bulk of the installation.
        install.run(self)

        # Build the BLAST databases now, as the user may not have write permissions to the
        # installation directory when running Kleborate.
        print('')
        data_dir = os.path.join(self.install_lib, 'kleborate', 'data')

        if not distutils.spawn.find_executable('makeblastdb'):
            print('Warning: could not find makeblastdb, so BLAST databases were not built.')
            print('They can be built when you run Kleborate, but this will require write '
                  'permissions to the data directory:')
            print(data_dir)
        else:
            check_dir_write_permission(data_dir)
            print('Building BLAST databases with makeblastdb:')
            try:
                for fasta in ['CARD_v3.1.13.fasta', 'clb_alleles.fasta', 'rmpA2.fasta',
                              'iro_alleles.fasta', 'iuc_alleles.fasta', 'rmp_alleles.fasta',
                              'Klebsiella_pneumoniae.fasta', 'wzi.fasta', 'ybt_alleles.fasta',
                              'MgrB_and_PmrB.fasta', 'OmpK.fasta', 'QRDR_120.fasta']:
                    build_blast_db(data_dir, fasta, 'nucl')
            except subprocess.CalledProcessError:
                print('\n')
                print('Warning: makeblastdb failed, so BLAST databases were not built.')
                print('They can be built when you run Kleborate, but this will require write '
                      'permissions to the data directory:')
                print(data_dir)

        ascii_art = " _   __ _     ______ ____   ____  _____       _______ ______ \n" \
                    "| | / /| |   |  ____|  _ \ / __ \|  __ \   /\|__   __|  ____|\n" \
                    "| |/ / | |   | |__  | |_) | |  | | |__) | /  \  | |  | |__   \n" \
                    "|   <  | |   |  __| |  _ <| |  | |  _  / / /\ \ | |  |  __|  \n" \
                    "| |\ \ | |___| |____| |_) | |__| | | \ \/ ____ \| |  | |____ \n" \
                    "|_| \_\|_____|______|____/ \____/|_|  \_\/    \_\_|  |______|\n"

        print('\n')
        print('\033[1m' + ascii_art + '\033[0m')  # bold formatting

        print('Kleborate is installed and ready to use!')
        print('')

        print('Example commands:')
        print('  kleborate --help')
        print('  kleborate -o results.txt -a *.fasta')
        print('  kleborate --resistance -o results.txt -a *.fasta')
        print('  kleborate --all -o results.txt -a *.fasta')
        print('')


setup(name='Kleborate',
      version=__version__,
      description='Kleborate',
      long_description=readme(),
      classifiers=['Development Status :: 4 - Beta',
                   'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 3',
                   'Topic :: Scientific/Engineering :: Bio-Informatics',
                   'Topic :: Scientific/Engineering :: Medical Science Apps.',
                   'Intended Audience :: Science/Research'],
      keywords='microbial genomics sequence typing',
      url='https://github.com/katholt/Kleborate',
      author='Kathryn Holt',
      author_email='',
      packages=['kleborate', 'kaptive'],
      install_requires=['biopython'],
      entry_points={'console_scripts': ['kleborate = kleborate.__main__:main']},
      include_package_data=True,
      zip_safe=False,
      cmdclass={'install': KleborateInstall})



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import os


def get_species_results(contigs, data_folder):
    species, species_hit_strength = get_klebsiella_species(contigs, data_folder)
    return {'species': species,
            'species_match': species_hit_strength}


def get_klebsiella_species(contigs, data_folder):
    f = os.popen('mash dist ' + data_folder + '/species_mash_sketches.msh ' + contigs)

    best_species = None
    best_distance = 1.0

    for line in f:
        line_parts = line.split('\t')
        reference = line_parts[0]
        if len(line_parts) < 4:
            continue
        species = reference.split('/')[0]
        distance = float(line_parts[2])

        # Fix up the species name formatting a bit.
        species = species.replace('Escherichia_coli', 'Escherichia coli / Shigella')
        species = species.replace('Raoultella', 'Klebsiella (Raoultella)')
        species = species.replace('_', ' ')
        species = species.replace(' subsp ', ' subsp. ')
        if species.endswith(' unknown'):
            species = species.replace(' unknown', ' (unknown species)')

        if distance < best_distance:
            best_distance = distance
            best_species = species

    f.close()

    if best_distance <= 0.02:
        return best_species, 'strong'
    elif best_distance <= 0.04:
        return best_species, 'weak'
    else:
        return 'unknown', ''


def is_kp_complex(results):
    """
    Returns True if the species call is in the Kp-complex, otherwise false.
    """
    assert 'species' in results
    species = results['species']
    if species.startswith('Klebsiella pneumoniae'):
        return True
    if species.startswith('Klebsiella quasipneumoniae'):
        return True
    if species.startswith('Klebsiella variicola'):
        return True
    if species.startswith('Klebsiella quasivariicola'):
        return True
    if species.startswith('Klebsiella africana'):
        return True
    return False



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import unittest

from kleborate.__main__ import get_rmp_mlst_results, get_rmpa2_results


class TestRmp(unittest.TestCase):
    """
    Tests RmpADC ST calls.
    """

    def setUp(self):
        self.data_dir = 'test/test_rmp/data'
        Args = collections.namedtuple('Args', ['min_coverage', 'min_identity',
                                               'min_spurious_coverage', 'min_spurious_identity'])
        self.args = Args(min_coverage=80.0, min_identity=90.0,
                         min_spurious_coverage=40.0, min_spurious_identity=80.0)

    def test_rmp_random(self):
        """
        This test has just random sequence and should give no rmp call.
        """
        results = get_rmp_mlst_results(self.data_dir, 'test/test_rmp/test_random.fasta', self.args)
        self.assertEqual(results['rmpA'], '-')
        self.assertEqual(results['rmpD'], '-')
        self.assertEqual(results['rmpC'], '-')
        self.assertEqual(results['RmpADC'], '-')
        self.assertEqual(results['RmST'], '0')

    def test_rmp_19(self):
        """
        This test is an exact match for RmST19, an ST without any truncated alleles.
        """
        results = get_rmp_mlst_results(self.data_dir, 'test/test_rmp/test_rmp_1.fasta', self.args)
        self.assertEqual(results['rmpA'], '7')
        self.assertEqual(results['rmpD'], '8')
        self.assertEqual(results['rmpC'], '3')
        self.assertEqual(results['RmpADC'], 'rmp 2A; VP')
        self.assertEqual(results['RmST'], '19')

    def test_rmp_21(self):
        """
        This test is an exact match for RmST21, an ST with a truncated allele.
        """
        results = get_rmp_mlst_results(self.data_dir, 'test/test_rmp/test_rmp_2.fasta', self.args)
        self.assertEqual(results['rmpA'], '7')
        self.assertEqual(results['rmpD'], '36')
        self.assertEqual(results['rmpC'], '4-17%')
        self.assertEqual(results['RmpADC'], 'rmp 2A; VP (truncated)')
        self.assertEqual(results['RmST'], '21')

    def test_rmp_novel(self):
        """
        This test is an exact match for alleles, but an unknown combination.
        """
        results = get_rmp_mlst_results(self.data_dir, 'test/test_rmp/test_rmp_3.fasta', self.args)
        self.assertEqual(results['rmpA'], '8')
        self.assertEqual(results['rmpD'], '1')
        self.assertEqual(results['rmpC'], '9-11%')
        self.assertEqual(results['RmpADC'], 'rmp unknown')
        self.assertEqual(results['RmST'], '0')

    def test_rmp_missing_1(self):
        """
        This test is a match for RmST1, except that it's missing rmpC.
        """
        results = get_rmp_mlst_results(self.data_dir, 'test/test_rmp/test_rmp_4.fasta', self.args)
        self.assertEqual(results['rmpA'], '41-54%')
        self.assertEqual(results['rmpD'], '20')
        self.assertEqual(results['rmpC'], '-')
        self.assertEqual(results['RmpADC'], 'rmp 2A; VP (incomplete) (truncated)')
        self.assertEqual(results['RmST'], '1-1LV')

    def test_rmpa2_random(self):
        """
        This test has just random sequence and should give no rmpA2 call.
        """
        results = get_rmpa2_results(self.data_dir, 'test/test_rmp/test_random.fasta', self.args)
        self.assertEqual(results['rmpA2'], '-')

    def test_rmpa2_exact(self):
        """
        This test has an exact match for rmpA2_1.
        """
        results = get_rmpa2_results(self.data_dir, 'test/test_rmp/test_rmpa2_1.fasta', self.args)
        self.assertEqual(results['rmpA2'], 'rmpA2_1')

    def test_rmpa2_exact_truncated(self):
        """
        This test has an exact match for rmpA2_7.
        """
        results = get_rmpa2_results(self.data_dir, 'test/test_rmp/test_rmpa2_2.fasta', self.args)
        self.assertEqual(results['rmpA2'], 'rmpA2_7-47%')

    def test_rmpa2_inexact(self):
        """
        This test has an inexact match for rmpA2_1.
        """
        results = get_rmpa2_results(self.data_dir, 'test/test_rmp/test_rmpa2_3.fasta', self.args)
        self.assertEqual(results['rmpA2'], 'rmpA2_1*')



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import tempfile
import os
import unittest

from kleborate.__main__ import get_output_headers, get_strain_name, get_contig_stat_results, \
    get_species_results, is_kp_complex, get_chromosome_mlst_results, get_ybt_mlst_results, \
    get_clb_mlst_results, get_iuc_mlst_results, get_iro_mlst_results, get_rmp_mlst_results,\
    get_rmpa2_results, get_wzi_and_k_locus_results, get_resistance_results, get_summary_results, \
    gunzip_contigs_if_necessary


def load_result_headers():
    with open('test/test_genomes/test_genome_results.txt', 'rt') as results:
        for line in results:
            parts = line.strip('\n').split('\t')
            assert parts[0] == 'strain'
            return parts


def load_results_one_genome(genome_name):
    with open('test/test_genomes/test_genome_results.txt', 'rt') as results:
        for line in results:
            parts = line.strip('\n').split('\t')
            if parts[0] == genome_name:
                return parts
    assert False


def load_correct_results(filename):
    genome_name = os.path.basename(filename)
    genome_name = genome_name.replace('.fasta.gz', '').replace('.fna.gz', '')
    headers = load_result_headers()
    results = load_results_one_genome(genome_name)
    assert len(headers) == len(results)
    return dict(zip(headers, results))


class TestGenomes(unittest.TestCase):

    def setUp(self):
        self.data_dir = 'test/test_genomes/data'
        Args = collections.namedtuple('Args', ['resistance', 'kaptive_k', 'kaptive_o',
                                               'min_coverage', 'min_identity'])
        self.args = Args(resistance=True, kaptive_k=False, kaptive_o=False,
                         min_coverage=80.0, min_identity=90.0)
        _, _, self.res_headers = get_output_headers(self.args, self.data_dir)

    def one_genome_test(self, filename):
        correct_results = load_correct_results(filename)
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(filename, tmp_dir)
            test_results = self.get_all_results(contigs)
        for result, value in correct_results.items():
            if value != '':
                assert test_results[result] == value

    def get_all_results(self, contigs):
        results = {'strain': get_strain_name(contigs)}
        results.update(get_species_results(contigs, self.data_dir))
        kp_complex = is_kp_complex(results)
        results.update(get_contig_stat_results(contigs, is_kp_complex))
        results.update(get_chromosome_mlst_results(self.data_dir, contigs, kp_complex, self.args))
        results.update(get_ybt_mlst_results(self.data_dir, contigs, self.args))
        results.update(get_clb_mlst_results(self.data_dir, contigs, self.args))
        results.update(get_iuc_mlst_results(self.data_dir, contigs, self.args))
        results.update(get_iro_mlst_results(self.data_dir, contigs, self.args))
        results.update(get_rmp_mlst_results(self.data_dir, contigs, self.args))
        results.update(get_rmpa2_results(self.data_dir, contigs, self.args))
        results.update(get_wzi_and_k_locus_results(self.data_dir, contigs, self.args))
        results.update(get_resistance_results(self.data_dir, contigs, self.args, self.res_headers,
                                              kp_complex))
        results.update(get_summary_results(results, self.res_headers))
        return results

    def test_172(self):
        self.one_genome_test('test/test_genomes/172.fasta.gz')

    def test_24042_7_219(self):
        self.one_genome_test('test/test_genomes/24042_7#219.fasta.gz')

    def test_24042_7_351(self):
        self.one_genome_test('test/test_genomes/24042_7#351.fasta.gz')

    def test_2685(self):
        self.one_genome_test('test/test_genomes/2685.fasta.gz')

    def test_AS7(self):
        self.one_genome_test('test/test_genomes/AS7.fna.gz')

    def test_BA4471(self):
        self.one_genome_test('test/test_genomes/BA4471.fasta.gz')

    def test_BA7971(self):
        self.one_genome_test('test/test_genomes/BA7971.fasta.gz')

    def test_CM3425(self):
        self.one_genome_test('test/test_genomes/CM3425.fasta.gz')

    def test_ERR2835900(self):
        self.one_genome_test('test/test_genomes/ERR2835900.fna.gz')

    def test_GCA_901563875(self):
        self.one_genome_test('test/test_genomes/GCA_901563875.1.fna.gz')

    def test_GCA_902158585(self):
        self.one_genome_test('test/test_genomes/GCA_902158585.1.fna.gz')

    def test_GCF_000009885(self):
        self.one_genome_test('test/test_genomes/GCF_000009885.1.fna.gz')

    def test_GCF_000016305(self):
        self.one_genome_test('test/test_genomes/GCF_000016305.1.fna.gz')

    def test_GCF_000019565(self):
        self.one_genome_test('test/test_genomes/GCF_000019565.1.fna.gz')

    def test_GCF_000215745(self):
        self.one_genome_test('test/test_genomes/GCF_000215745.1.fna.gz')

    def test_GCF_000240325(self):
        self.one_genome_test('test/test_genomes/GCF_000240325.1.fna.gz')

    def test_GCF_000247855(self):
        self.one_genome_test('test/test_genomes/GCF_000247855.1.fna.gz')

    def test_GCF_000247895(self):
        self.one_genome_test('test/test_genomes/GCF_000247895.1.fna.gz')

    def test_GCF_000492415(self):
        self.one_genome_test('test/test_genomes/GCF_000492415.1.fna.gz')

    def test_GCF_000492795(self):
        self.one_genome_test('test/test_genomes/GCF_000492795.1.fna.gz')

    def test_GCF_000523395(self):
        self.one_genome_test('test/test_genomes/GCF_000523395.1.fna.gz')

    def test_GCF_000648315(self):
        self.one_genome_test('test/test_genomes/GCF_000648315.1.fna.gz')

    def test_GCF_000733495(self):
        self.one_genome_test('test/test_genomes/GCF_000733495.1.fna.gz')

    def test_GCF_000829965(self):
        self.one_genome_test('test/test_genomes/GCF_000829965.1.fna.gz')

    def test_GCF_000968155(self):
        self.one_genome_test('test/test_genomes/GCF_000968155.1.fna.gz')

    def test_GCF_001068035(self):
        self.one_genome_test('test/test_genomes/GCF_001068035.1.fna.gz')

    def test_GCF_002108345(self):
        self.one_genome_test('test/test_genomes/GCF_002108345.1.fna.gz')

    def test_GCF_002247645(self):
        self.one_genome_test('test/test_genomes/GCF_002247645.1.fna.gz')

    def test_GCF_002248955(self):
        self.one_genome_test('test/test_genomes/GCF_002248955.1.fna.gz')

    def test_GCF_002806645(self):
        self.one_genome_test('test/test_genomes/GCF_002806645.1.fna.gz')

    def test_GCF_003095495(self):
        self.one_genome_test('test/test_genomes/GCF_003095495.1.fna.gz')

    def test_GCF_003345475(self):
        self.one_genome_test('test/test_genomes/GCF_003345475.1.fna.gz')

    def test_GCF_003400925(self):
        self.one_genome_test('test/test_genomes/GCF_003400925.1.fna.gz')

    def test_GCF_003937345(self):
        self.one_genome_test('test/test_genomes/GCF_003937345.1.fna.gz')

    def test_GCF_004010735(self):
        self.one_genome_test('test/test_genomes/GCF_004010735.1.fna.gz')

    def test_GCF_005860775(self):
        self.one_genome_test('test/test_genomes/GCF_005860775.1.fna.gz')

    def test_GCF_900501255(self):
        self.one_genome_test('test/test_genomes/GCF_900501255.1.fna.gz')

































    # def test_GCF_900501255(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_900501255.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '0')
    #         self.assertEqual(results['rmpA'], 'rmpA_2(KpVP-1)')
    #         self.assertEqual(results['rmpA2'], 'rmpA2_4*-50%')
    #         self.assertEqual(results['AGly'], '-')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], '-')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], '-')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], 'SHV-187*')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_003400925(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_003400925.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '5')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], 'rmpA2_9')
    #         self.assertEqual(results['AGly'], 'RmtC')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'GyrA-83I;ParC-80I')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], 'SulI')
    #         self.assertEqual(results['Tet'], 'TetA')
    #         self.assertEqual(results['Tmt'], 'DfrA1')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], '-')
    #         self.assertEqual(results['Bla_Carb'], 'NDM-1')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], 'SHV-11')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_001068035(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_001068035.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '0')
    #         self.assertEqual(results['rmpA'], 'rmpA_2(KpVP-1)')
    #         self.assertEqual(results['rmpA2'], 'rmpA2_3*-0%')
    #         self.assertEqual(results['AGly'], '-')
    #
    #         # PmrB is split in two: 40% (start) and 60% (end). The correct answer is 40% as that
    #         # includes the beginning of the gene.
    #         self.assertEqual(results['Col'], 'PmrB-40%')
    #
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], '-')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], '-')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], '-')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_003095495(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_003095495.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '17')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], 'Aac3-IId^;AadA2^;Aph3-Ia^;RmtB;Sat-2A;StrA^;StrB')
    #
    #         # PmrB is split in two: 36% (start) and 66% (end). The correct answer is 36% as that
    #         # includes the beginning of the gene.
    #         self.assertEqual(results['Col'], 'MgrB-62%;PmrB-36%')
    #
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'GyrA-83I;ParC-80I')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], 'Erm42*;MphA')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], 'CatA1^')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], 'SulI;SulII')
    #         self.assertEqual(results['Tet'], 'TetG')
    #         self.assertEqual(results['Tmt'], 'DfrA12?')
    #         self.assertEqual(results['Omp'], 'OmpK35-25%;OmpK36GD')
    #         self.assertEqual(results['Bla'], 'TEM-1D^')
    #         self.assertEqual(results['Bla_Carb'], 'KPC-2')
    #         self.assertEqual(results['Bla_ESBL'], 'CTX-M-14')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], 'SHV-11')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_000009885(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_000009885.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '0')
    #         self.assertEqual(results['rmpA'], 'rmpA_11(ICEKp1),rmpA_2(KpVP-1)')
    #         self.assertEqual(results['rmpA2'], 'rmpA2_3-47%')
    #         self.assertEqual(results['AGly'], '-')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], '-')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], '-')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], '-')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], 'SHV-11^')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #         self.assertEqual(results['wzi'], 'wzi1')
    #         self.assertEqual(results['K_locus'], 'KL1')
    #
    # def test_GCF_004010735(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_004010735.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['wzi'], '-')
    #         self.assertEqual(results['K_locus'], '-')
    #
    # def test_GCF_000968155(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_000968155.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '0')
    #         self.assertEqual(results['rmpA'], 'rmpA_9(KpVP-2)')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], '-')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], '-')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], '-')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], '-')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_002108345(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_002108345.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '11')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], 'Aac3-IId^;Aph3-Ia^')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'GyrA-83I;ParC-80I;QnrB4;QnrB4')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], 'CatA2*;CatB3*')
    #         self.assertEqual(results['Rif'], 'Arr3')
    #         self.assertEqual(results['Sul'], 'SulI*;SulI^')
    #         self.assertEqual(results['Tet'], '-')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], 'OmpK35-18%')
    #         self.assertEqual(results['Bla'], 'DHA-1;OXA-1;SHV-187*')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_002247645(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_002247645.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '7')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], 'Aac3-IId^')
    #         self.assertEqual(results['Col'], '-')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'Qnr-S1')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], 'FloR*')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], 'TetA*')
    #         self.assertEqual(results['Tgc'], 'TetX4')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], 'LAP-2;SHV-2a^;TEM-1D^')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_002248955(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_002248955.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '4')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], 'Aac3-IId^')
    #         self.assertEqual(results['Col'], 'Mcr3-1*')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'GyrA-83F;GyrA-87A;ParC-80I')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], 'CatA1^')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], '-')
    #         self.assertEqual(results['Tet'], 'TetA')
    #         self.assertEqual(results['Tmt'], '-')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], 'SHV-28^')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_GCF_003345475(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/GCF_003345475.1.fna.gz',
    #                                               tmp_dir)
    #         results = self.get_all_results(contigs)
    #         self.assertEqual(results['num_resistance_genes'], '5')
    #         self.assertEqual(results['rmpA'], '-')
    #         self.assertEqual(results['rmpA2'], '-')
    #         self.assertEqual(results['AGly'], '-')
    #         self.assertEqual(results['Col'], 'Mcr1-1;MgrB-49%')
    #         self.assertEqual(results['Fcyn'], '-')
    #         self.assertEqual(results['Flq'], 'Qnr-S1')
    #         self.assertEqual(results['Gly'], '-')
    #         self.assertEqual(results['MLS'], '-')
    #         self.assertEqual(results['Ntmdz'], '-')
    #         self.assertEqual(results['Phe'], '-')
    #         self.assertEqual(results['Rif'], '-')
    #         self.assertEqual(results['Sul'], 'SulI')
    #         self.assertEqual(results['Tet'], 'TetA')
    #         self.assertEqual(results['Tmt'], 'DfrA1')
    #         self.assertEqual(results['Omp'], '-')
    #         self.assertEqual(results['Bla'], 'SHV-110')
    #         self.assertEqual(results['Bla_Carb'], '-')
    #         self.assertEqual(results['Bla_ESBL'], '-')
    #         self.assertEqual(results['Bla_ESBL_inhR'], '-')
    #         self.assertEqual(results['Bla_broad'], '-')
    #         self.assertEqual(results['Bla_broad_inhR'], '-')
    #
    # def test_AS7(self):
    #     with tempfile.TemporaryDirectory() as tmp_dir:
    #         contigs = gunzip_contigs_if_necessary('test/test_genomes/AS7.fna.gz', tmp_dir)
    #         results = self.get_all_results(contigs)
    #
    #         # This genome is missing the start of rmpA, which counts as a truncation to 0%.
    #         self.assertEqual(results['rmpA'], 'rmpA_1*-0%(KpVP-1)')




"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import unittest

from kleborate.contig_stats import get_contig_stats, get_qc_warnings


class TestContigStats(unittest.TestCase):

    def test_count_1(self):
        contig_count, _, _, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_1.fasta')
        self.assertEqual(contig_count, 4)

    def test_count_2(self):
        contig_count, _, _, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_2.fasta')
        self.assertEqual(contig_count, 3)

    def test_n50_1(self):
        _, n50, _, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_1.fasta')
        self.assertEqual(n50, 40)

    def test_n50_2(self):
        _, n50, _, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_2.fasta')
        self.assertEqual(n50, 200)

    def test_longest_1(self):
        _, _, longest_contig, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_1.fasta')
        self.assertEqual(longest_contig, 45)

    def test_longest_2(self):
        _, _, longest_contig, _, _ = get_contig_stats('test/test_contig_stats/contig_stats_2.fasta')
        self.assertEqual(longest_contig, 200)

    def test_ambiguous_bases_1(self):
        _, _, _, _, ambiguous = get_contig_stats('test/test_contig_stats/contig_stats_1.fasta')
        self.assertEqual(ambiguous, 'no')

    def test_ambiguous_bases_2(self):
        _, _, _, _, ambiguous = get_contig_stats('test/test_contig_stats/contig_stats_2.fasta')
        self.assertEqual(ambiguous, 'yes (1)')

    def test_ambiguous_bases_3(self):
        _, _, _, _, ambiguous = get_contig_stats('test/test_contig_stats/contig_stats_3.fasta')
        self.assertEqual(ambiguous, 'no')

    def test_ambiguous_bases_4(self):
        _, _, _, _, ambiguous = get_contig_stats('test/test_contig_stats/contig_stats_4.fasta')
        self.assertEqual(ambiguous, 'yes (4)')

    def test_total_size_1(self):
        _, _, _, total_size, _ = get_contig_stats('test/test_contig_stats/contig_stats_1.fasta')
        self.assertEqual(total_size, 115)

    def test_total_size_2(self):
        _, _, _, total_size, _ = get_contig_stats('test/test_contig_stats/contig_stats_2.fasta')
        self.assertEqual(total_size, 260)

    def test_total_size_3(self):
        _, _, _, total_size, _ = get_contig_stats('test/test_contig_stats/contig_stats_3.fasta')
        self.assertEqual(total_size, 260)

    def test_total_size_4(self):
        _, _, _, total_size, _ = get_contig_stats('test/test_contig_stats/contig_stats_4.fasta')
        self.assertEqual(total_size, 260)

    def test_qc_warnings_1(self):
        # A perfectly nice assembly - yields no warnings.
        warnings = get_qc_warnings(5500000, 250000, 'no', True)
        self.assertEqual(warnings, '-')

    def test_qc_warnings_2(self):
        # Large assembly size.
        warnings = get_qc_warnings(10000000, 250000, 'no', True)
        self.assertEqual(warnings, 'total_size')

    def test_qc_warnings_3(self):
        # Large assembly size, but not KpSC so no warning.
        warnings = get_qc_warnings(10000000, 250000, 'no', False)
        self.assertEqual(warnings, '-')

    def test_qc_warnings_4(self):
        # Small assembly size.
        warnings = get_qc_warnings(2000000, 250000, 'no', True)
        self.assertEqual(warnings, 'total_size')

    def test_qc_warnings_5(self):
        # Small assembly size, but not KpSC so no warning.
        warnings = get_qc_warnings(2000000, 250000, 'no', False)
        self.assertEqual(warnings, '-')

    def test_qc_warnings_6(self):
        # Small N50.
        warnings = get_qc_warnings(5500000, 1000, 'no', True)
        self.assertEqual(warnings, 'N50')

    def test_qc_warnings_7(self):
        # Has ambiguous bases
        warnings = get_qc_warnings(5500000, 250000, 'yes (50)', True)
        self.assertEqual(warnings, 'ambiguous_bases')

    def test_qc_warnings_8(self):
        # All three warnings.
        warnings = get_qc_warnings(2000000, 1000, 'yes (1000)', True)
        self.assertEqual(warnings, 'total_size,N50,ambiguous_bases')



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections
import tempfile
import unittest

from kleborate.__main__ import get_chromosome_mlst_results, gunzip_contigs_if_necessary, \
    get_kp_subspecies_based_on_st


class TestMlst(unittest.TestCase):
    """
    Tests chromosomal MLST calls.
    """

    def setUp(self):
        self.data_dir = 'test/test_mlst/data'
        Args = collections.namedtuple('Args', ['min_coverage', 'min_identity'])
        self.args = Args(min_coverage=80.0, min_identity=90.0)

    def test_chromosome_random(self):
        """
        This test has just random sequence and should give no MLST call.
        """
        results = get_chromosome_mlst_results(self.data_dir,
                                              'test/test_mlst/test_random.fasta', True, self.args)
        self.assertEqual(results['gapA'], '-')
        self.assertEqual(results['infB'], '-')
        self.assertEqual(results['mdh'], '-')
        self.assertEqual(results['pgi'], '-')
        self.assertEqual(results['phoE'], '-')
        self.assertEqual(results['rpoB'], '-')
        self.assertEqual(results['tonB'], '-')
        self.assertEqual(results['ST'], '0')
        self.assertEqual(results['Chr_ST'], '0')

    def test_chromosome_exact(self):
        """
        This test is an exact match for ST23.
        """
        results = get_chromosome_mlst_results(self.data_dir,
                                              'test/test_mlst/test_mlst_1.fasta', True, self.args)
        self.assertEqual(results['gapA'], '2')
        self.assertEqual(results['infB'], '1')
        self.assertEqual(results['mdh'], '1')
        self.assertEqual(results['pgi'], '1')
        self.assertEqual(results['phoE'], '9')
        self.assertEqual(results['rpoB'], '4')
        self.assertEqual(results['tonB'], '12')
        self.assertEqual(results['ST'], 'ST23')
        self.assertEqual(results['Chr_ST'], 'ST23')

    def test_chromosome_inexact(self):
        """
        This test is is one base off from ST23 (single substitution in mdh_1 of G to A).
        """
        results = get_chromosome_mlst_results(self.data_dir,
                                              'test/test_mlst/test_mlst_2.fasta', True, self.args)
        self.assertEqual(results['gapA'], '2')
        self.assertEqual(results['infB'], '1')
        self.assertEqual(results['mdh'], '1*')
        self.assertEqual(results['pgi'], '1')
        self.assertEqual(results['phoE'], '9')
        self.assertEqual(results['rpoB'], '4')
        self.assertEqual(results['tonB'], '12')
        self.assertEqual(results['ST'], 'ST23-1LV')
        self.assertEqual(results['Chr_ST'], 'ST23-1LV')

    def test_unknown_ST(self):
        """
        This test has exact matches for alleles but in an unknown ST combination
        """
        results = get_chromosome_mlst_results(self.data_dir,
                                              'test/test_mlst/test_mlst_3.fasta', True, self.args)
        self.assertEqual(results['gapA'], '44')
        self.assertEqual(results['infB'], '56')
        self.assertEqual(results['mdh'], '34')
        self.assertEqual(results['pgi'], '19')
        self.assertEqual(results['phoE'], '30')
        self.assertEqual(results['rpoB'], '53')
        self.assertEqual(results['tonB'], '55')
        self.assertEqual(results['ST'], '0')
        self.assertEqual(results['Chr_ST'], '0')

    def test_83(self):
        contigs = 'test/test_mlst/83.fasta.gz'
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)
            results = get_chromosome_mlst_results(self.data_dir, contigs, True, self.args)
            self.assertEqual(results['gapA'], '10')
            self.assertEqual(results['infB'], '7')
            self.assertEqual(results['mdh'], '1')
            self.assertEqual(results['pgi'], '1')
            self.assertEqual(results['phoE'], '1')
            self.assertEqual(results['rpoB'], '1')
            self.assertEqual(results['tonB'], '35')
            self.assertEqual(results['ST'], 'ST160')
            self.assertEqual(results['Chr_ST'], 'ST160')

    def test_134(self):
        contigs = 'test/test_mlst/134.fasta.gz'
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)
            results = get_chromosome_mlst_results(self.data_dir, contigs, True, self.args)
            self.assertEqual(results['gapA'], '2')
            self.assertEqual(results['infB'], '1')
            self.assertEqual(results['mdh'], '2')
            self.assertEqual(results['pgi'], '1')
            self.assertEqual(results['phoE'], '4')
            self.assertEqual(results['rpoB'], '4')
            self.assertEqual(results['tonB'], '4')
            self.assertEqual(results['ST'], 'ST16')
            self.assertEqual(results['Chr_ST'], 'ST16')

    def test_ba779(self):
        contigs = 'test/test_mlst/BA779.fasta.gz'
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)
            results = get_chromosome_mlst_results(self.data_dir, contigs, True, self.args)
            self.assertEqual(results['gapA'], '2')
            self.assertEqual(results['infB'], '1')
            self.assertEqual(results['mdh'], '1')
            self.assertEqual(results['pgi'], '1')
            self.assertEqual(results['phoE'], '9')
            self.assertEqual(results['rpoB'], '4')
            self.assertEqual(results['tonB'], '12')
            self.assertEqual(results['ST'], 'ST23')
            self.assertEqual(results['Chr_ST'], 'ST23')

    def test_ozaenae(self):
        contigs = 'test/test_mlst/GCF_900451425.1.fna.gz'
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)
            results = get_chromosome_mlst_results(self.data_dir, contigs, True, self.args)
            self.assertEqual(results['ST'], 'ST90 (subsp. ozaenae)')
            self.assertEqual(results['Chr_ST'], 'ST90')

    def test_rhinoscleromatis(self):
        contigs = 'test/test_mlst/GCF_000163455.1.fna.gz'
        with tempfile.TemporaryDirectory() as tmp_dir:
            contigs = gunzip_contigs_if_necessary(contigs, tmp_dir)
            results = get_chromosome_mlst_results(self.data_dir, contigs, True, self.args)
            self.assertEqual(results['ST'], 'ST67 (subsp. rhinoscleromatis)')
            self.assertEqual(results['Chr_ST'], 'ST67')

    def test_kp_subspecies_1(self):
        # Just test some non-ozaenae non-rhinoscleromatis STs.
        self.assertEqual(get_kp_subspecies_based_on_st('ST1'), 'ST1')
        self.assertEqual(get_kp_subspecies_based_on_st('ST12'), 'ST12')
        self.assertEqual(get_kp_subspecies_based_on_st('ST123'), 'ST123')
        self.assertEqual(get_kp_subspecies_based_on_st('ST1234'), 'ST1234')
        self.assertEqual(get_kp_subspecies_based_on_st('ST1-1LV'), 'ST1-1LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST12-1LV'), 'ST12-1LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST123-1LV'), 'ST123-1LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST1234-1LV'), 'ST1234-1LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST1-2LV'), 'ST1-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST12-2LV'), 'ST12-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST123-2LV'), 'ST123-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST1234-2LV'), 'ST1234-2LV')

    def test_kp_subspecies_2(self):
        # Test the ozaenae STs.
        self.assertEqual(get_kp_subspecies_based_on_st('ST90'), 'ST90 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST91'), 'ST91 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST92'), 'ST92 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST93'), 'ST93 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST95'), 'ST95 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST96'), 'ST96 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST97'), 'ST97 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST381'), 'ST381 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST777'), 'ST777 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3193'), 'ST3193 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3766'), 'ST3766 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3768'), 'ST3768 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3771'), 'ST3771 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3781'), 'ST3781 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3782'), 'ST3782 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3784'), 'ST3784 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3802'), 'ST3802 (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3803'), 'ST3803 (subsp. ozaenae)')

    def test_kp_subspecies_3(self):
        # Test the ozaenae STs with 1LV - should still result in ozaenae.
        self.assertEqual(get_kp_subspecies_based_on_st('ST90-1LV'), 'ST90-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST91-1LV'), 'ST91-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST92-1LV'), 'ST92-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST93-1LV'), 'ST93-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST95-1LV'), 'ST95-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST96-1LV'), 'ST96-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST97-1LV'), 'ST97-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST381-1LV'), 'ST381-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST777-1LV'), 'ST777-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3193-1LV'), 'ST3193-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3766-1LV'), 'ST3766-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3768-1LV'), 'ST3768-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3771-1LV'), 'ST3771-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3781-1LV'), 'ST3781-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3782-1LV'), 'ST3782-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3784-1LV'), 'ST3784-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3802-1LV'), 'ST3802-1LV (subsp. ozaenae)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3803-1LV'), 'ST3803-1LV (subsp. ozaenae)')

    def test_kp_subspecies_4(self):
        # Test the ozaenae STs with 2LV - should no longer result in ozaenae.
        self.assertEqual(get_kp_subspecies_based_on_st('ST90-2LV'), 'ST90-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST91-2LV'), 'ST91-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST92-2LV'), 'ST92-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST93-2LV'), 'ST93-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST95-2LV'), 'ST95-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST96-2LV'), 'ST96-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST97-2LV'), 'ST97-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST381-2LV'), 'ST381-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST777-2LV'), 'ST777-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3193-2LV'), 'ST3193-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3766-2LV'), 'ST3766-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3768-2LV'), 'ST3768-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3771-2LV'), 'ST3771-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3781-2LV'), 'ST3781-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3782-2LV'), 'ST3782-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3784-2LV'), 'ST3784-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3802-2LV'), 'ST3802-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3803-2LV'), 'ST3803-2LV')

    def test_kp_subspecies_5(self):
        # Test the rhinoscleromatis STs.
        self.assertEqual(get_kp_subspecies_based_on_st('ST67'), 'ST67 (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST68'), 'ST68 (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST69'), 'ST69 (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3772'),
                         'ST3772 (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3819'),
                         'ST3819 (subsp. rhinoscleromatis)')

    def test_kp_subspecies_6(self):
        # Test the rhinoscleromatis STs with 1LV - should still result in rhinoscleromatis.
        self.assertEqual(get_kp_subspecies_based_on_st('ST67-1LV'),
                         'ST67-1LV (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST68-1LV'),
                         'ST68-1LV (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST69-1LV'),
                         'ST69-1LV (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3772-1LV'),
                         'ST3772-1LV (subsp. rhinoscleromatis)')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3819-1LV'),
                         'ST3819-1LV (subsp. rhinoscleromatis)')

    def test_kp_subspecies_7(self):
        # Test the rhinoscleromatis STs with 2LV - should no longer result in rhinoscleromatis.
        self.assertEqual(get_kp_subspecies_based_on_st('ST67-2LV'), 'ST67-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST68-2LV'), 'ST68-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST69-2LV'), 'ST69-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3772-2LV'), 'ST3772-2LV')
        self.assertEqual(get_kp_subspecies_based_on_st('ST3819-2LV'), 'ST3819-2LV')



"""
Copyright 2020 Kat Holt
Copyright 2020 Ryan Wick (rrwick@gmail.com)
https://github.com/katholt/Kleborate/

This file is part of Kleborate. Kleborate is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Kleborate is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Kleborate. If
not, see <http://www.gnu.org/licenses/>.
"""

import collections

from .misc import load_fasta


def get_contig_stat_results(contigs, kp_complex):
    contig_count, n50, longest_contig, total_size, ambiguous_bases = get_contig_stats(contigs)
    qc_warnings = get_qc_warnings(total_size, n50, ambiguous_bases, kp_complex)
    return {'contig_count': str(contig_count),
            'N50': str(n50),
            'largest_contig': str(longest_contig),
            'total_size': str(total_size),
            'ambiguous_bases': ambiguous_bases,
            'QC_warnings': qc_warnings}


def get_contig_stats(assembly):
    fasta = load_fasta(assembly)

    base_counts = collections.defaultdict(int)
    for _, seq in fasta:
        for b in seq:
            base_counts[b] += 1
    base_counts.pop('A', None)
    base_counts.pop('C', None)
    base_counts.pop('G', None)
    base_counts.pop('T', None)
    ambiguous_base_count = sum(base_counts.values())
    if ambiguous_base_count:
        ambiguous_bases = 'yes (' + str(ambiguous_base_count) + ')'
    else:
        ambiguous_bases = 'no'

    contig_lengths = sorted([len(x[1]) for x in fasta])
    if not contig_lengths:
        return 0, 0, 0
    longest = contig_lengths[-1]

    total_size = sum(contig_lengths)
    half_total_length = total_size / 2
    total_so_far = 0
    segment_lengths = contig_lengths[::-1]
    for length in segment_lengths:
        total_so_far += length
        if total_so_far >= half_total_length:
            n50 = length
            break
    else:
        n50 = 0

    return len(contig_lengths), n50, longest, total_size, ambiguous_bases


def get_qc_warnings(total_size, n50, ambiguous_bases, kp_complex):
    warnings = []

    # Genome size should be 5 to 6.5 Mbp for KpSC genomes. For non-KpSC genomes, we don't check
    # the genome size.
    if kp_complex:
        if total_size < 5000000:
            warnings.append('total_size')
        elif total_size > 6500000:
            warnings.append('total_size')

    if n50 < 10000:
        warnings.append('N50')

    if 'yes' in ambiguous_bases:
        warnings.append('ambiguous_bases')

    if warnings:
        return ','.join(warnings)
    else:
        return '-'
